# encoding: utf-8
"""

"""
__author__ = 'Richard Smith'
__date__ = '08 Jun 2021'
__copyright__ = 'Copyright 2018 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'richard.d.smith@stfc.ac.uk'

import argparse
import yaml
from pydoc import locate
import pkg_resources

from asset_scanner.core import BaseExtractor
from asset_scanner.core.utils import load_plugins
from asset_scanner.core.exceptions import NoPluginsError

import logging


def command_args():
    """
    Sets the command line arguments and handles their parsing
    :return: command line options
    """
    parser = argparse.ArgumentParser(description='Run the asset scanner as configured')
    parser.add_argument('conf', help='Path to a yaml configuration file')
    args = parser.parse_args()

    return args


def setup_logging(conf):
    config = conf.get('logging', {})
    if not config or (config and not config.get('format')):
        config['format'] = '%(asctime)s %(name)s %(levelname)s %(message)s'

    logging.basicConfig(**config)


def load_config(path):
    with open(path) as reader:
        conf = yaml.safe_load(reader)
    return conf


def load_extractor(conf: dict) -> BaseExtractor:
    """
    Load the extractor.

    Looks for extractor defined in the configuration in preference
    and falls back to the first defined entry point at ``asset_scanner.extractors``

    :param conf: Configuration dict
    :return: Extractor
    """

    extractor = None

    if conf.get('extractor'):
        extractor = locate(conf['extractor'])
        if not extractor:
            raise ImportError(f'Unable to find {conf["extractor"]}. Check that it is installed.')

    if not extractor:
        for entry_point in pkg_resources.iter_entry_points('asset_scanner.extractors'):
            extractor = entry_point.load()

            # Only load the first one
            break

    if not extractor:
        raise NoPluginsError('No extraction plugins have been loaded')

    return extractor(conf)


def main():
    args = command_args()

    conf = load_config(args.conf)

    setup_logging(conf)

    extractor = load_extractor(conf)

    input_plugins = load_plugins(conf, 'asset_scanner.input_plugins', 'inputs')

    for input in input_plugins:
        input.run(extractor)


if __name__ == '__main__':
    main()



# encoding: utf-8
"""
Base Extractor
--------------

This module provides the base class for all derived extractors.

"""
__author__ = 'Richard Smith'
__date__ = '08 Jun 2021'
__copyright__ = 'Copyright 2018 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'richard.d.smith@stfc.ac.uk'

from abc import ABC, abstractmethod
from .utils import load_plugins
from .handler_picker import HandlerPicker
from .item_describer import ItemDescriptions, ItemDescription
from asset_scanner.types.source_media import StorageType
import re

from typing import List


class BaseExtractor(ABC):
    """
    Base class to define an interface for other extractor classes

    Attributes:

        PROCESSOR_ENTRY_POINT:
            Defines the entry point to look for in the setup.py for the downstream
            package. This is used by the ``asset_scanner`` command to load the extractor
            if the extractor is not specifically defined in the the configuration file.
    """
    PROCESSOR_ENTRY_POINT = None

    def __init__(self, conf: dict):
        self.conf = conf
        self.processors = self.load_processors()
        self.output_plugins = self.load_output_plugins()
        self.item_descriptions = ItemDescriptions(conf['item_descriptions']['root_directory']) if 'item_descriptions' in conf else None

        self.load_processors()

    @staticmethod
    def _get_path(filepath, **kwargs) -> str:
        """
        Check to see if there is a `ParseResult <https://docs.python.org/3/library/urllib.parse.html?highlight=parseresult#urllib.parse.ParseResult>_
        object. If there is and it contains a network location, return just the path
        e.g.
        - https://data.ceda.ac.uk/badc/ukmo -> /badc/ukmo
        - http://cmip6-zarr-o.s3.jc.rl.ac.uk/CMIP6.CFMIP.IPSL.IPSL-CM6A-LR/amip-p4K.r1i1p1f1.Amon.evspsbl.gr.v20180906.zarr -> CMIP6.CFMIP.IPSL.IPSL-CM6A-LR/amip-p4K.r1i1p1f1.Amon.evspsbl.gr.v20180906.zarr

        :param filepath:
        :param kwargs:
        :return:
        """
        parse_result = kwargs.get('uri_parse')
        if not parse_result:
            return filepath

        if not parse_result.netloc:
            return filepath

        return parse_result.path

    @staticmethod
    def _get_category(string, label, regex):
        """

        :param string:
        :param label:
        :param regex:
        :return:

        """

        m = re.search(regex, string)

        if not m:
            label = None

        return label

    def get_categories(self, filepath: str, source_media: StorageType, description: ItemDescription) -> List:
        """
        Get asset category labels

        :param filepath: Asset path
        :param source_media: Source media class
        :param description: ItemDescription for asset
        :return:

        """
        categories = set()

        for conf in description.categories:
            label = self._get_category(filepath, **conf)
            if label:
                categories.add(label)

        return list(categories) or ['data']

    def load_processors(self, entrypoint: str = None) -> HandlerPicker:
        return HandlerPicker(entrypoint or self.PROCESSOR_ENTRY_POINT)

    def load_output_plugins(self) -> List:
        return load_plugins(self.conf, 'asset_scanner.output_plugins', 'outputs')

    @abstractmethod
    def process_file(self, filepath: str, source_media: StorageType = StorageType.POSIX, **kwargs) -> None:
        pass

    def output(self, filepath: str, source_media: StorageType, data: dict, namespace: str = None) -> None:
        for backend in self.output_plugins:
            if not backend.namespace or backend.namespace == namespace:
                backend.export(data)



# encoding: utf-8
"""

"""
__author__ = 'Richard Smith'
__date__ = '02 Jun 2021'
__copyright__ = 'Copyright 2018 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'richard.d.smith@stfc.ac.uk'

from abc import ABC, abstractmethod
from asset_scanner.core import BaseExtractor
from asset_scanner.core.utils import load_plugins
from asset_scanner.types.source_media import StorageType


class BaseInputPlugin(ABC):

    def __init__(self, **kwargs):
        self.filters = None
        if kwargs.get('filters'):
            self.filters = load_plugins(kwargs, 'asset_scanner.plugin_filters', 'filters')

    def should_process(self, filepath, source_media: StorageType) -> bool:
        """
        Should the path be sent for processing?

        Will run through any filter plugins. All plugins must pass for a True
        response. Any False will short circuit the logic and return False

        :param filepath: Filepath to test
        :param source_media: Source media

        :return: Bool, ``default: True``
        """
        if self.filters:
            return any((filter.run(filepath, source_media) for filter in self.filters))

        return True

    @abstractmethod
    def run(self, extractor: BaseExtractor):
        ...



# encoding: utf-8
"""
Object Store Input
------------------

Takes an endpoint url and optionally a bucket prefix and delimiter and will
scan the items at these points in the object store, submitting each to the
asset extractor

**Plugin name:** ``object_store``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``session_kwargs``
      - ``dict``
      - ``REQUIRED`` Dictionary containing the S3 access and secret keys
    * - ``endpoint_url``
      - ``string``
      - ``REQUIRED`` URL for S3 endpoint
    * - ``bucket``
      - ``string``
      - Bucket to be scanned. If ``None`` all buckets in endpoint will be scanned.
    * - ``prefix``
      - ``string``
      - Only items with prefix will be scanned
    * - ``delimiter``
      - ``string``
      - Group items after delimiter into one object

Example Configuration:
    .. code-block:: yaml

        inputs:
            - name: object_store
                endpoint_url: https://cedadev-o.s3-ext.jc.rl.ac.uk
                session_kwargs: {
                    aws_access_key_id: ACCESS_KEY,
                    aws_secret_access_key: SECRET_KEY
                }
                bucket: my_bucket
                prefix: directory_or_file
                delimiter: .zarr/

"""
__author__ = 'Rhys Evans'
__date__ = '02 Jun 2021'
__copyright__ = 'Copyright 2018 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'rhys.r.evans@stfc.ac.uk'


from .base import BaseInputPlugin

import boto3

from typing import TYPE_CHECKING
import logging

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from asset_scanner.core import BaseExtractor


class ObjectStoreInputPlugin(BaseInputPlugin):
    """
    """

    def __init__(self, **kwargs):

        self.endpoint_url = kwargs.get('endpoint_url')
        session = boto3.session.Session(**kwargs['session_kwargs'])
        s3 = session.resource(
            's3',
            endpoint_url=self.endpoint_url,
        )

        self.client = s3.meta.client
        self.buckets = [s3.Bucket(kwargs['bucket'])] if 'bucket' in kwargs else s3.buckets.all()
        self.prefix = kwargs.get('prefix', '')
        self.delimiter = kwargs.get('delimiter', '')

    def run(self, extractor: 'BaseExtractor' ):

        for bucket in self.buckets:
            total_files = 0
            for obj in bucket.objects.filter(Prefix=self.prefix, Delimiter=self.delimiter):
                extractor.process_file(f'{self.endpoint_url}/{bucket.name}/{obj.key}', 'object_store', client=self.client)
                total_files += 1

            LOGGER.info(f'Processed {total_files} files from {bucket}')



# encoding: utf-8
"""
Elasticsearch
-------------

An output backend which outputs the content generated to elasticsearch
using the Elasticsearch API

**Plugin name:** ``elasticsearch``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``connection_kwargs``
      - ``dict``
      - ``REQUIRED`` Connection kwargs passed to the `elasticsearch client  <https://elasticsearch-py.readthedocs.io/en/latest/api.html#elasticsearch>`_
    * - ``index.name``
      - ``str``
      - ``REQUIRED`` The index to output the content.
    * - ``index.mapping``
      - ``str``
      - Path to a yaml file which defines the mapping for the index
    * - ``namespace``
      - ``str``
      - Can be used by downstream processors to separate outputs to different indices or clusters

Example Configuration:
    .. code-block:: yaml

        outputs:
            - name: elasticsearch
              connection_kwargs:
                hosts: ['host1','host2']
              index:
                name: 'assets-2021-06-02'
              namespace: assets
"""
__author__ = 'Richard Smith'
__date__ = '01 Jun 2021'
__copyright__ = 'Copyright 2018 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'richard.d.smith@stfc.ac.uk'

from .base import OutputBackend
from elasticsearch import Elasticsearch
from asset_scanner.core.utils import load_yaml, Coordinates

from typing import Dict


class ElasticsearchOutputBackend(OutputBackend):
    """
    Connects to an elasticsearch instance and exports the
    documents to elasticsearch.

    """

    CLEAN_METHODS = [
            '_format_bbox',
            '_format_temporal_extent'
        ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        index_conf = kwargs['index']

        self.es = Elasticsearch(**kwargs['connection_kwargs'])
        self.index_name = index_conf['name']
        self.pipeline_conf = kwargs.get('pipeline', None)

        # Create the index, if it doesn't already exist
        if index_conf.get('mapping'):
            if not self.es.indices.exists(self.index_name):
                mapping = load_yaml(index_conf.get('mapping'))
                self.es.indices.create(self.index_name, body=mapping)

    @staticmethod
    def _format_bbox(data: Dict) -> Dict:
        """
        Convert WGS84 coordinates into GeoJSON and
        format for Elasticsearch. Replaces the bbox key.

        :param data: Input data dictionary
        """
        body = data['body']

        if body.get('bbox'):
            bbox = body.pop('bbox')

            body['spatial'] = {
                    "bbox": {
                        "type": "envelope",
                        "coordinates": Coordinates.from_wgs84(
                            bbox).to_geojson()
                    }
                }

        return data

    @staticmethod
    def _format_temporal_extent(data: Dict) -> Dict:
        """
        Convert `extent object<https://github.com/radiantearth/stac-spec/blob/master/collection-spec/collection-spec.md#extent-object>_` for Elasticsearch.

        :param data: Input data dictionary
        """
        body = data['body']

        if body.get('extent', {}).get('temporal'):
            temporal_extent = body['extent'].pop('temporal')

            if temporal_extent[0][0]:
                body['extent']['temporal'] = {'gte': temporal_extent[0][0]}
            if temporal_extent[0][1]:
                temporal = body['extent'].get('temporal',{})
                temporal['lte'] = temporal_extent[0][1]
                body['extent']['temporal'] = temporal

        return data

    def clean(self, data: Dict) -> Dict:
        """
        Condition the input dictionary for elasticsearch
        :param data: Input dictionary
        :returns: Dictionary produced as a result of the clean methods
        """

        for method in self.CLEAN_METHODS:
            m = getattr(self, method)
            data = m(data)

        return data

    def export(self, data, **kwargs):

        data = self.clean(data)

        index_kwargs = {
            'index': self.index_name,
            'id': data['id'],
            'body': {
                'doc': data['body'],
                'doc_as_upsert': True
            }
        }

        self.es.update(**index_kwargs)



# encoding: utf-8
"""
Exceptions
----------

"""
__author__ = 'Richard Smith'
__date__ = '02 Jun 2021'
__copyright__ = 'Copyright 2018 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'richard.d.smith@stfc.ac.uk'


class NoPluginsError(Exception):
    """
    Raised when no plugins have been loaded for a particular section.
    i.e. if no input plugins were successfully loaded, then this exception
    would be raised.
    """
    pass


# encoding: utf-8
"""

"""
__author__ = 'Richard Smith'
__date__ = '23 Sep 2021'
__copyright__ = 'Copyright 2018 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'richard.d.smith@stfc.ac.uk'


# encoding: utf-8
"""

"""
__author__ = 'Richard Smith'
__date__ = '01 Jun 2021'
__copyright__ = 'Copyright 2018 United Kingdom Research and Innovation'
__license__ = 'BSD - see LICENSE file in top-level package directory'
__contact__ = 'richard.d.smith@stfc.ac.uk'

from abc import ABC, abstractmethod


class OutputBackend(ABC):

    def __init__(self, namespace=None, **kwargs):
        self.namespace = namespace

    @abstractmethod
    def export(self, data, **kwargs):
        pass


from setuptools import setup, find_packages

with open("README.md") as readme_file:
    _long_description = readme_file.read()

setup(
    name='asset_scanner',
    version='0.1.1',
    description='Framework to provide plugin architecture to allow the scanning of assets to extract metadata and facets.',
    author='Richard Smith',
    url='https://github.com/cedadev/asset-scanner/',
    long_description=_long_description,
    long_description_content_type='text/markdown',
    license='BSD - See asset_extractor/LICENSE file for details',
    packages=find_packages(),
    test_suite='tests',
    package_data={
        'asset_scanner': [
            'LICENSE'
        ]
    },
    install_requires=[
        'pyyaml',
        'ceda-directory-tree',
        'tqdm'
    ],
    extras_require={
        'docs': [
            'sphinx',
            'sphinx-rtd-theme',
            'elasticsearch',
            'sphinxcontrib-programoutput'
        ],
        'elasticsearch': [
            'elasticsearch'
        ],
        'intake-esm': [
            'intake-esm'
        ],
        'rabbitmq': [
            'pika'
        ]
    },
    python_requires='>=3.5',
    tests_require=[
        'pytest'
    ],
    entry_points={
        'console_scripts': [
            'asset_scanner = asset_scanner.scripts.asset_scanner:main',
        ],
        'asset_scanner.output_plugins': [
            'standard_out = asset_scanner.plugins.output_plugins.standard_out:StdoutOutputBackend',
            'elasticsearch = asset_scanner.plugins.output_plugins.elasticsearch_backend:ElasticsearchOutputBackend',
        ],
        'asset_scanner.plugin_filters': [
            'path_regex = asset_scanner.plugins.filters.path_regex:PathRegexFilter',
        ],
        'asset_scanner.input_plugins': [
            'file_system = asset_scanner.plugins.input_plugins.file_system_input:FileSystemInputPlugin',
            'object_store = asset_scanner.plugins.input_plugins.object_store_input:ObjectStoreInputPlugin',
            'intake_esm = asset_scanner.plugins.input_plugins.intake_esm_input:IntakeESMInputPlugin',
            'rabbit_mq = asset_scanner.plugins.input_plugins.rabbit_mq_input:RabbitMQInputPlugin',
        ]
    }
)
